#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# pyright: reportUnboundVariable=false
# pyright: reportOperatorIssue=false

# import external libraries
import numpy as np  # computation with matrices, vectors
import matplotlib.pyplot as plt  # plot curves
import pandas as pd   # to extract data from REW txt files
import mod_fdm_exciter as fdmexciter  # call to module mod_fdm_exciter.py
from scipy.signal import find_peaks  # peaks detection in impedance
from scipy.optimize import curve_fit  # curve fiiting function

def read_data(file_path, X, Y, Z):
    """
    Read 3 columns X, Y, Z from a text file file_path (ie frequency, impedance, phase or frequency, SPL, phase from REW
    file_path = file path to the file to read
    X, Y, Z = column names
    """
    data = pd.read_csv(file_path, sep='\\s+', header=None, names=[X, Y, Z])
    return data

def secondorder_model2(f, *params):
    """
    Define the model function for the curve fitting
    f = frequency array (similar as x values)
    model = model array output (similar to y values)
    params = [Q0, k, f0, then all the values of the impedance peaks considered]
    """
    model = np.zeros_like(f)  # prepare an output array
    admittance = np.zeros_like(f, dtype=complex)  # the main loop of the function works on complex
    omega = 2*np.pi*f  # amgular frequency
    # Q model
    Q0 = params[0]
    k = params[1]
    f0 = params[2]
    Q = Q0 * (1 + (f/f0)**k)
    # loop to compute the admittance over the provided impedance peaks (start at index 3 of params)
    for i in range(3, len(params)):
        omega0 = 2*np.pi*peak_freqs[i-3]  # WARNING : use of a global variable peak_freq[]
        # Note :
        # When curve fit will call this functions, it will adjust the values in params to fit at the best the provided target values
        # So the frequency peaks are in a global variable to avoid curve fit changes these values
        A = params[i]
        admittance += A * 1j * omega / omega0 / (1 + 1j * omega / omega0 / Q - (omega/omega0)**2) / Q
    model = np.absolute(admittance)
    return model

# Enter here the path to the REW file
# Note : for Windows PC, adapt the path structure
path = '../data/part3/'
# zfile = 'Z_PS2CCCC.txt'
zfile = 'Z_depron9_60x26.txt'
zfile_path = path + zfile
print(zfile_path)

# Calculation parameter
WITH_RMS = False # Set to True to substract the exciter Rms, False to get a global Q, exciter damping included

# Exciter model
EXTMODEL = "DAEX25FHE4"  # Choose an exciter available in the exciter library
# Exciter parameters
BL, RE, LE, MMS, CMS, RMS, XR, XI, KR, KI, FV, NV = fdmexciter.get_exciter(EXTMODEL)

# Parameter init for curve fitting. Always better to start from the estimated value for the material
Q_INIT = 40
K_INIT = 0
F0_INIT = 10000

# Extract Impedance Target file (ie from REW)
data = read_data(zfile_path, 'Frequency', 'Impedance', 'Phase')
FREQZREW = data['Frequency'].to_numpy()
ZMODREW = data['Impedance'].to_numpy()
ZPHIREW = data['Phase'].to_numpy()
ZREW = ZMODREW*(np.cos(np.pi*ZPHIREW/180) + 1j * np.sin(np.pi*ZPHIREW/180))

# Compute admittance
RMIN = np.mean(np.absolute(ZREW[FREQZREW < 30])) # autodetection of RE (avoid RE error)
ZEE = RMIN + KR*(2*np.pi*FREQZREW)**XR + 1j*KI*(2*np.pi*FREQZREW)**XI
YM = (ZREW - ZEE)/BL**2
if WITH_RMS:
    YT = (ZREW - ZEE)/BL**2
    YM = 1/(1/YT - RMS*(1 + (FV/FREQZREW)**NV))
YM_MODULE = np.absolute(YM)

# Find admittance peaks
# Note : adjust height and/or prominence to get not to many peaks
ympeaks, _ = find_peaks(YM_MODULE[FREQZREW < 2000], height=0.07, prominence=0.03)
peak_freqs = FREQZREW[ympeaks]
peak_amps = YM_MODULE[ympeaks]

Q = np.zeros(len(ympeaks))

# plot impedance from REW and electrical part of the model (ZEE)
fig = plt.figure(figsize=(10, 8))  # create a figure
plt.semilogx(FREQZREW, np.absolute(ZREW), label='|Ze| measurement')
plt.semilogx(FREQZREW, np.absolute(ZEE), label='|Re + Xe| model')
plt.legend()
plt.title(zfile)
plt.xlabel('frequency Hz')
plt.ylabel('electrical impedance Ohm')
plt.show()

# plot admittance and targeted peaks
fig = plt.figure(figsize=(10, 8))  # create a figure
plt.semilogx(FREQZREW, YM_MODULE, label='|ym|')
plt.semilogx(FREQZREW[ympeaks], np.absolute(YM[ympeaks]), 'rx', label='peaks to fit')
plt.legend()
plt.xlabel('frequency Hz')
plt.ylabel('mechanichal admittance m/Ns')
plt.title(zfile)
plt.show()

# Curve fitting init ----------------------------------------
p0 = []  # is provided to curve fit function as initialisation
p0.append(Q_INIT)
p0.append(K_INIT)
p0.append(F0_INIT)
for A in peak_amps:
    p0.append(A)

# Curve fitting parameters boundaries -----------------------
bound_low = []  # curve fit won't propose parameters lower than
bound_low.append(5)  # Q
bound_low.append(0)  # k
bound_low.append(10)  # f0
for A in peak_amps:
    bound_low.append(A/2)  # peak value
bound_high = []   # curve fit won't propose parameters higher than
bound_high.append(200)  # Q
bound_high.append(3)  # k
bound_high.append(20000)  # f0
for A in peak_amps:
    bound_high.append(2*A) # peak value
bounds = (bound_low, bound_high)  # boundaries for curve fit

# --------------- Curve fitting --------------------------
# Here is the magic of the Scipy function
# the adjusted parameters are in popt
# not use here : an estimation of the confidence in the provided parameters can be extracted from pcov
popt, pcov = curve_fit(secondorder_model2, FREQZREW, YM_MODULE, p0=p0, bounds=bounds, method='trf', maxfev=20000)
# --------------------------------------------------------

# Results
ym_model = secondorder_model2(FREQZREW, *popt)  # let's compute the model admittance and its main parameters below
q_model = popt[0]
k_model = popt[1]
f0_model = popt[2]
print(f"Curve fitting results :")
print(f"With Rms = {WITH_RMS}")
print(f"Q0 = {q_model:.3f}")
print(f"k = {k_model:.3f}")
print(f"f0 = {f0_model:.3f}")
print(f"Q@100Hz = {q_model*(1+(f0_model/100)**k_model):.3f}")
# This an error estimation in order to compare different optimization
error = np.sqrt(np.sum((YM_MODULE[(FREQZREW < 2000) & (FREQZREW > 20)] - ym_model[(FREQZREW < 2000) & (FREQZREW > 20)])**2))
print(f"error = {error:.3f}")

# Result plot
# Final plot : REW admittance, model admittance, Q model
fig, ax1 = plt.subplots(figsize=(10, 6))
ax1.semilogx(FREQZREW, YM_MODULE, color='blue', label='|ym| Measurement')
plt.semilogx(FREQZREW, ym_model, color='green', label='|ym| Model')
ax1.set_ylabel('mechanichal admittance |ym|', color='blue')
ax1.tick_params('y', colors='blue')
ax2 = ax1.twinx()
ax2.semilogx(FREQZREW, q_model * (1+(FREQZREW/f0_model)**k_model), color='red', label='Q Model')
ax2.set_ylabel('Q factor', color='red')
ax2.tick_params('y', colors='red')
lines1, labels1 = ax1.get_legend_handles_labels()
lines2, labels2 = ax2.get_legend_handles_labels()
ax1.legend(lines1 + lines2, labels1 + labels2, loc='upper left')
ax1.set_xlabel('frequency Hz')
fig.suptitle(zfile)
plt.show()
